# -*- coding: utf-8 -*-
# Author: Simone Marsili <simomarsili@gmail.com>
# License: BSD 3 clause
"""Utility functions."""

import functools
import logging

logger = logging.getLogger(__name__)

__all__ = [
    'is_command',
]


def is_command(cmds):
    """Given one command returns its path, or None.
    Given a list of commands returns the first recoverable path, or None.
    """
    try:
        from shutil import which  # python3 only
    except ImportError:
        from distutils.spawn import find_executable as which

    if isinstance(cmds, str):
        return which(cmds)
    for cmd in cmds:
        path = which(cmd)
        if path is not None:
            return path
    return path


def open_tempfile():
    """Open a temp file."""
    import tempfile
    tempfile = tempfile.NamedTemporaryFile
    kwargs = {'delete': True, 'mode': 'r+'}
    return tempfile(**kwargs)


def timeit(func):
    """Timeit decorator."""

    @functools.wraps(func)
    def timed(*args, **kwargs):
        import time
        ts0 = time.time()
        result = func(*args, **kwargs)
        ts1 = time.time()
        logging.debug('%r: %2.4f secs', func, ts1 - ts0)
        return result

    return timed



# -*- coding: utf-8 -*-
# Author: Simone Marsili <simomarsili@gmail.com>
# License: BSD 3 clause
"""alignment parser"""
import logging
import sys

import gopen

import lilbio

logger = logging.getLogger(__name__)


class LilBioError(Exception):
    """Base class for lilbio exceptions."""


class UnsupportedFormatError(LilBioError):
    """Unsupported format."""


class NotCallableError(LilBioError):
    """Argument func did not receive a callable object."""


def parse(source, fmt, func=None):
    """
    Parse an alignment file into records.

    Parameters
    ----------
    source : {str, file object}
        Alignment filename of a file handle.
    fmt : str
        Format of the alignment file.
        Valid values are: {'fasta', 'stockholm'}
    func : callable or iterable
        When passed, apply to the sequence string and return a string or a list
        of one-letter codes.
        If `func` is an iterable of functions, use the corresponding composite
        function e.g. if func == [f, g]: func = lambda x: g(f(x)).

    Yields
    ------
    (identifier, seq) : tuple (str, str)
        For each record, a tuple containing identifier and sequence.

    """
    from collections import Iterable
    from lilbio.bioparsers import bioparsers
    from lilbio.funcs import compose

    try:
        bioparser = bioparsers[fmt]
    except KeyError:
        raise UnsupportedFormatError(fmt)

    if func is None:

        def func(x):  # pylint: disable=function-redefined
            return x

    if isinstance(func, Iterable):
        func = compose(*func)

    if not callable(func):
        raise NotCallableError(func)

    data = gopen.gread(source)
    parsed_data = bioparser(data)
    for title, seq in parsed_data:
        yield title, func(seq)


def write(a, f=sys.stdout):
    """Write as FASTA file."""
    tostr = lilbio.funcs.tostr
    for r in a:
        string_record = '>%s\n%s' % (r[0], tostr(r[1]))
        print(string_record, file=f)


def main():
    """main."""
    import argparse
    parser_ = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser_.add_argument('-i', '--infile', type=str, help='Input file name')
    parser_.add_argument('-f',
                         '--format',
                         type=str,
                         help='Input format',
                         default='fasta')
    args = parser_.parse_args()
    filename = args.infile
    fmt = args.format

    parsed_records = parse(filename, fmt)
    for identifier, seq in parsed_records:
        print('>%s\n%s' % (identifier, seq))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Author: Simone Marsili <simomarsili@gmail.com>
# License: BSD 3 clause
"""
Functions of sequences. If a transformed version of the sequence is
returned, it should be a list of symbols (not a string).
"""
import functools
import logging

logger = logging.getLogger(__name__)
# uppercase plus *,- symbols
GENERIC_ALPHABET = set('*-ABCDEFGHIJKLMNOPQRSTUVWXYZ')
# Canonical/natural codes plus gap symbol.
PROTEIN_ALPHABET = set('-ACDEFGHIKLMNPQRSTVWY')


def compose(*funcs):
    """Composition of functions from iterable.

    Functions operate in iteration order:
    compose([a, b])(x) == b(a(x)).
    """

    def gf(f, g):
        """Compose two functions."""
        return lambda x: g(f(x))

    return functools.reduce(gf, funcs, lambda x: x)


def from_alphabet_only(s, alphabet):
    """Filter symbols from alphabet."""
    return [c for c in s if c in alphabet]


def uppercase_only(s):
    """Filter uppercase + '-' symbols."""
    return from_alphabet_only(s, GENERIC_ALPHABET)


def tostr(s):
    """Stringify sequence s."""
    return ''.join(s)



# -*- coding: utf-8 -*-
# Copyright 2006-2015 by Peter Cock.  All rights reserved.
# Revisions copyright 2015 by Ben Woodcroft.  All rights reserved.
# Modifications copyright 2016 by Simone Marsili.
#
# Biopython License Agreement
# ---------------------------
#
# Permission to use, copy, modify, and distribute this software and its
# documentation with or without modifications and for any purpose and
# without fee is hereby granted, provided that any copyright notices
# appear in all copies and that both those copyright notices and this
# permission notice appear in supporting documentation, and that the
# names of the contributors or copyright holders not be used in
# advertising or publicity pertaining to distribution of the software
# without specific prior permission.

# THE CONTRIBUTORS AND COPYRIGHT HOLDERS OF THIS SOFTWARE DISCLAIM ALL
# WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE
# CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
# OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
# OR PERFORMANCE OF THIS SOFTWARE.
"""(Decorated) Biopython parsers."""
import logging

logger = logging.getLogger(__name__)
__all__ = [
    'bioparsers',
    'fasta_parser',
    'stockholm_parser',
]

bioparsers = {}


def bioparser(parsers, fmt=None):
    """Register the format of the decorated parser generator function.
    """

    def decorator(parser_func):
        import inspect
        try:
            assert inspect.isgeneratorfunction(parser_func)
        except AssertionError:
            raise TypeError('%r is not a valid parser generator function')
        if fmt:
            fmt1 = str(fmt)
        else:
            fmt1 = parser_func.__name__

        def decorated_parser(source):
            """Handle iterables, file-like and compressed files."""
            import collections
            import gopen
            if isinstance(source, collections.Iterable):
                for name, seq in parser_func(source):
                    yield name, seq
            else:
                with gopen.readable(source) as f:
                    for name, seq in parser_func(f):
                        yield name, seq

        parsers[fmt1] = decorated_parser

        return decorated_parser

    return decorator


@bioparser(bioparsers, 'fasta')
def fasta_parser(fileo):
    """Generator of Fasta records.
    Adapted from Biopython SimpleFastaParser.
    (https://raw.githubusercontent.com/biopython/biopython/biopython-168/Bio/SeqIO/FastaIO.py)

    For each record a tuple of two strings is returned, the FASTA title
    line (without the leading '>' character), and the sequence (with any
    whitespace removed). The title line is not divided up into an
    identifier (the first word) and comment or description.

    Parameters
    ----------
    fileo : file object

    Returns
    -------
    (seqname, sequence) : (str, str)
        Next record in the alignment as tuple (seqname, sequence)

    """
    # Skip any text before the first record (e.g. blank lines, comments)
    # TODO: add a maximum number of lines for the initial comment

    # make sure that fileo is an iterator
    fileo = iter(fileo)
    # skip blank lines (only) till a record is found
    while True:
        line = next(fileo, None)
        if not line:
            return  # Premature end of file, or just empty?
        if line[0] == '>':
            break
        elif line[0] != ';':
            raise ValueError('Not Fasta format')

    while True:
        if line[0] != '>':
            raise ValueError(
                'Records in Fasta files should start with ">" character')
        title = line[1:].rstrip()
        lines = []
        line = next(fileo, None)
        while line:
            if line[0] == '>':
                break
            lines.append(line.rstrip())
            line = next(fileo, None)

        # Remove trailing whitespace, and any internal spaces
        # (and any embedded \r which are possible in mangled files
        # when not opened in universal read lines mode)
        yield (title, ''.join(lines).replace(' ', '').replace('\r', ''))

        if not line:
            return  # StopIteration

    assert False, 'Should not reach this line'


@bioparser(bioparsers, 'stockholm')
def stockholm_parser(fileo):
    """Generator of Stockholm records.
    Adapted from Biopyhon StockholmIterator.
    (https://raw.githubusercontent.com/biopython/biopython/biopython-168/Bio/AlignIO/StockholmIO.py)

    Parameters
    ----------
    fileo : file object

    Returns
    -------
    (seqname, sequence) : (str, str)
        Next record in the alignment as tuple (seqname, sequence)

    """

    # make sure that fileo is an iterator
    fileo = iter(fileo)
    while True:
        line = next(fileo, None)
        if not line:
            break  # end of file
        line = line.strip()  # remove trailing \n
        if line == '//':
            return  # end of alignment
        if line == '':
            # blank line, ignore
            pass
        elif line[0] != '#':
            if line[0] in ['>', ';']:
                raise ValueError('Not Stockholm format')
            # Sequence
            # Format: '<seqname> <sequence>'
            parts = [x.strip() for x in line.split(' ', 1)]
            if len(parts) != 2:
                # This might be someone attempting to store
                # a zero length sequence?
                raise ValueError('Could not split line into identifier '
                                 'and sequence:\n' + line)
            title, seq = parts
            # do not replace dots with gaps!
            # seqs[id] += seq.replace(".", "-")
            yield (title, seq)

    assert False, 'Should not reach this line'



# -*- coding: utf-8 -*-
"""Setup module."""
from os import path

from setuptools import setup

# from setuptools import find_packages

NAME = 'little-bio-parser'
PACKAGES = ['lilbio']
PACKAGE_FILE = 'package.json'
SETUP_REQUIRES = []
INSTALL_REQUIRES = [
    # 'gopen @ http://github.com/simomarsili/gopen/archive/v0.3.1.tar.gz']
    # 'gopen @ http://github.com/simomarsili/gopen/archive/master.tar.gz'
    'gopen>=0.6'
]
EXTRAS_REQUIRES = {'test': ['pytest']}


def get_version(source):
    """ Retrieve version number."""
    import json
    with open(source, 'r') as fp:
        version_data = json.load(fp)
    try:
        return version_data['version']
    except KeyError:
        # no version number in package.json
        raise KeyError('check version file: no version number')


def get_long_description(here):
    """Get the long description from the README file."""
    import codecs
    with codecs.open(path.join(here, 'README.rst'), encoding='utf-8') as _rf:
        return _rf.read()


HERE = path.abspath(path.dirname(__file__))
VERSION = get_version(path.join(HERE, PACKAGE_FILE))
LONG_DESCRIPTION = get_long_description(HERE)

setup(
    name=NAME,
    version=VERSION,
    description='A template project with packages',
    long_description=LONG_DESCRIPTION,
    author='Simone Marsili',
    author_email='simo.marsili@gmail.com',
    url='https://github.com/simomarsili/' + NAME,
    packages=PACKAGES,
    # packages=find_packages(exclude=['tests']),
    package_data={
        '': ['LICENSE.txt', 'README.rst', 'requirements.txt', 'package.json']
    },
    include_package_data=True,
    setup_requires=SETUP_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRES,
    license='BSD 3-Clause',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



# -*- coding: utf-8 -*-
# Author: Simone Marsili <simomarsili@gmail.com>
# License: BSD 3 clause
"""A little parser for alignments of biological sequences."""

import pkg_resources

from lilbio.funcs import uppercase_only
from lilbio.parser import parse, write

project_name = 'little-bio-parser'
__version__ = pkg_resources.require(project_name)[0].version
__copyright__ = 'Copyright (C) 2017 Simone Marsili'
__license__ = 'BSD 3 clause'
__author__ = 'Simone Marsili (simo.marsili@gmail.com)'
__all__ = ['parse', 'write', 'uppercase_only']
